# coding=utf8

import json

from datetime import timedelta, datetime as dt
from context import get_json_value
from util import print_traceback, dumpjs, format_datetime, AssetStatus
from easydb_api import easydb_search


def get_expired_timestamps(min_age_days):
    ts = dt.utcnow()
    if min_age_days == 0:
        return format_datetime(ts)
    ts = ts - timedelta(days=min_age_days)
    return format_datetime(ts)


def do_search(easydb_context, easydb_server, easydb_token, query):
    logger = easydb_context.get_logger('pf.plugin.base.auto_keyworder.search')
    try:
        logger.debug('query:\n%s' % dumpjs(query))
        statuscode, result = easydb_search(easydb_server,
                                           easydb_token,
                                           query)
        logger.debug('result:\n%s' % dumpjs(result))
        if statuscode != 200:
            raise Exception('[%s] %s' % (str(statuscode), result))
        return result
    except Exception as e:
        print_traceback(logger, e)
        return None


def format_search_from_tagfilter(tagfilter):

    if tagfilter is None:
        return None

    sub_searches = []

    if len(tagfilter.any) > 0:
        sub_searches.append(
            {
                'bool': 'should',
                'fields': [
                    '_tags._id'
                ],
                'in': tagfilter.any,
                'type': 'in'
            }
        )

    if len(tagfilter.no) > 0:
        sub_searches.append(
            {
                'bool': 'must_not',
                'fields': [
                        '_tags._id'
                ],
                'in': tagfilter.no,
                'type': 'in'
            }
        )

    if len(tagfilter.all) > 0:
        sub_searches.append({
            'bool': 'must',
            'type': 'complex',
            'search': [
                {
                    'bool': 'must',
                    'fields': [
                        '_tags._id'
                    ],
                    'in': [
                        tag
                    ],
                    'type': 'in'
                }
                for tag in tagfilter.all
            ]
        })

    if len(sub_searches) < 1:
        return None

    return {
        'bool': 'must',
        'search': sub_searches,
        'type': 'complex'
    }


def format_search_for_objecttype(objecttype):
    return {
        'bool': 'must',
        'fields': [
                '_objecttype'
        ],
        'in': [
            objecttype
        ],
        'type': 'in'
    }


def build_linked_object_search_query(objecttype, fieldname, keywords, language=None, offset=0, limit=1000):
    field = '%s.%s' % (objecttype, fieldname)

    searches = [
        format_search_for_objecttype(objecttype),
        {
            'bool': 'must',
            'fields': [
                field
            ],
            'in': keywords,
            'type': 'in'
        }
    ]

    query = {
        'format': 'short',
        'generate_rights': False,
        'offset': offset,
        'limit': limit,
        'fields': [
            {
                'key': 'id',
                'field': '%s._id' % (objecttype)
            },
            {
                'key': 'keyword',
                'field': field
            }
        ],
        'objecttypes': [
            objecttype
        ],
        'search': searches
    }

    if language is not None:
        query['language'] = language

    return query


def build_search_query(objecttype, tagfilter, asset_field, timestamp_field, min_age_days=0, offset=0, limit=1000):

    searches = [
        format_search_for_objecttype(objecttype),
        {
            'bool': 'must_not',
            'fields': [
                '%s.%s' % (objecttype, asset_field)
            ],
            'in': [
                None
            ],
            'type': 'in'
        },
        {
            "type": "in",
            "fields": [
                '%s.%s.class' % (objecttype, asset_field)
            ],
            "in": [
                "image"
            ],
            "bool": "must"
        }
    ]

    tf_search = format_search_from_tagfilter(tagfilter)
    if tf_search is not None:
        searches.append(tf_search)

    if min_age_days is not None:
        searches.append({
            'bool': 'must',
            'search': [
                {
                    'bool': 'should',
                    'field': '%s.%s' % (objecttype, timestamp_field),
                    'to': get_expired_timestamps(min_age_days),
                    'type': 'range'
                },
                {
                    'bool': 'should',
                    'fields': [
                        '%s.%s' % (objecttype, timestamp_field)
                    ],
                    'in': [
                        None
                    ],
                    'type': 'in'
                }
            ],
            'type': 'complex'
        })

    query = {
        'format': 'long',
        'generate_rights': False,
        'offset': offset,
        'limit': limit,
        'exclude_fields': [
            '_standard',
            '_format',
            '_created',
            '_last_modified',
            '_published_count',
            '_collections',
            '_uuid',
            '_global_object_id'
        ],
        'objecttypes': [
            objecttype
        ],
        'search': searches
    }

    return query


def map_search_results(result, objecttype, object_map):
    objects = get_json_value(result, 'objects')
    if not isinstance(objects, list):
        raise Exception('invalid response: %s' % dumpjs(result))

    n = 0
    for o in objects:
        _sys_id = get_json_value(o, '_system_object_id', True)
        if _sys_id in object_map:
            continue

        _id = get_json_value(o, '%s._id' % objecttype, True)
        mask = get_json_value(o, '_mask', True)
        tags = get_json_value(o, '_tags', True)

        object_map[_sys_id] = AssetStatus({
            '_mask': mask,
            '_tags': tags,
            '_objecttype': objecttype,
            '_system_object_id': _sys_id,
            objecttype: get_json_value(o, objecttype, True)
        })

        n += 1

    return n
